# --------------------------------------------------------------------------------------
# Copyright (c) 2021-2022, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
import collections.abc
from typing import Any, ClassVar, MutableMapping, Type

from .catom import Member
from .dict import Dict as ADict
from .instance import Instance
from .list import List as AList
from .scalars import Bool, Bytes, Callable as ACallable, Float, Int, Str, Value
from .set import Set as ASet
from .subclass import Subclass
from .typing_utils import extract_types, get_args, is_optional

_NO_DEFAULT = object()

_TYPE_TO_MEMBER = {
    bool: Bool,
    int: Int,
    float: Float,
    str: Str,
    bytes: Bytes,
    list: AList,
    dict: ADict,
    set: ASet,
    collections.abc.Callable: ACallable,
}


def generate_member_from_type_or_generic(
    type_generic: Any, default: Any, annotate_type_containers: int
) -> Member:
    """Generate a member from a type or generic alias."""
    types = extract_types(type_generic)
    parameters = get_args(type_generic)

    m_kwargs = {}

    m_cls: Type[Member]
    if any(
        isinstance(t, type) and issubclass(t, Member) for t in types
    ) and not isinstance(default, Member):
        raise ValueError(
            "Member subclasses cannot be used as annotations without "
            "specifying a default value for the attribute."
        )
    elif object in types or Any in types:
        m_cls = Value
    # Int, Float, Str, Bytes, List, Dict, Set, Tuple, Bool, Callable
    elif len(types) == 1 and types[0] in _TYPE_TO_MEMBER:
        t = types[0]
        m_cls = _TYPE_TO_MEMBER[t]
        if annotate_type_containers and t in (list, dict, set, tuple):
            # We can only validate homogeneous tuple so far so we ignore other cases
            if t is tuple:
                if (...) in parameters or len(set(parameters)) == 1:
                    parameters = (parameters[0],)
                else:
                    parameters = ()
            parameters = tuple(
                generate_member_from_type_or_generic(
                    t, _NO_DEFAULT, annotate_type_containers - 1
                )
                if t not in (Any, object)
                else None
                for t in parameters
            )
        else:
            parameters = ()

    # The value was annotated with Type[T] so we use a subclass
    elif all(t is type for t in types):
        m_cls = Subclass
    else:
        # We cannot determine if a type has a trivial __instancecheck__ or not so
        # we always use Instance since Typed will fail to validate with generic types
        # such as collections.Iterable
        m_cls = Instance

        opt, filtered_types = is_optional(types)
        parameters = (filtered_types,)
        m_kwargs["optional"] = opt
        if opt and default not in (_NO_DEFAULT, None):
            raise ValueError(
                "Members requiring Instance(optional=True) cannot have "
                "a non-None default value."
            )
        elif not opt and default is not _NO_DEFAULT:
            raise ValueError("Members requiring Instance cannot have a default value.")

        # Instance does not have a default keyword so turn a None default into the
        # equivalent no default.
        default = _NO_DEFAULT

    if default is not _NO_DEFAULT:
        m_kwargs["default"] = default

    return m_cls(*parameters, **m_kwargs)


def generate_members_from_cls_namespace(
    cls_name: str, namespace: MutableMapping[str, Any], annotate_type_containers: int
) -> None:
    """Generate the member corresponding to a type annotation."""
    annotations = namespace["__annotations__"]
    from .atom import set_default

    for name, ann in annotations.items():
        default = namespace.get(name, _NO_DEFAULT)

        # We skip field for which a member was already provided or annotations
        # corresponding to class variables.
        if isinstance(default, (Member, set_default)):

            # Allow string annotations for members
            if isinstance(ann, str):
                continue

            types = extract_types(ann)
            if len(types) != 1 or not issubclass(types[0], Member):
                raise TypeError(
                    f"Field '{name}' of '{cls_name}' is assigned a Member-like value "
                    "but its annotation is not Member compatible"
                )
            continue

        # We also skip fields annotated as class variables.
        elif getattr(ann, "__origin__", None) is ClassVar:
            continue

        try:
            namespace[name] = generate_member_from_type_or_generic(
                ann, default, annotate_type_containers
            )
        except ValueError as e:
            raise ValueError(
                "Encountered an issue when generating a member for field "
                f"'{name}' of '{cls_name}'."
            ) from e



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Callable, Dict, Optional, Tuple, Type, TypeVar, Union, overload

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

class Subclass(Member[T, T]):
    # No default
    @overload
    def __new__(cls, kind: Type[T], default: None = None) -> Subclass[Type[T]]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T]], default: None = None
    ) -> Subclass[Type[T]]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: None = None
    ) -> Subclass[Union[Type[T], Type[T1]]]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1], Type[T2]], default: None = None
    ) -> Subclass[Union[Type[T], Type[T1], Type[T2]]]: ...
    # With default
    @overload
    def __new__(cls, kind: Type[T], default: Type[T]) -> Subclass[Type[T]]: ...
    @overload
    def __new__(cls, kind: Tuple[Type[T]], default: Type[T]) -> Subclass[Type[T]]: ...
    @overload
    def __new__(
        cls, kind: Tuple[Type[T], Type[T1]], default: Union[Type[T], Type[T1]]
    ) -> Subclass[Union[Type[T], Type[T1]]]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        default: Union[Type[T], Type[T1], Type[T2]],
    ) -> Subclass[Union[Type[T], Type[T1], Type[T2]]]: ...

class ForwardSubclass(Subclass[T]):
    @overload
    def __new__(cls, resolve: Callable[[], Type[T]]) -> ForwardSubclass[Type[T]]: ...
    @overload
    def __new__(
        cls, resolve: Callable[[], Tuple[Type[T]]]
    ) -> ForwardSubclass[Type[T]]: ...
    @overload
    def __new__(
        cls, resolve: Callable[[], Tuple[Type[T], Type[T1]]]
    ) -> ForwardSubclass[Union[Type[T], Type[T1]]]: ...
    @overload
    def __new__(
        cls, resolve: Callable[[], Tuple[Type[T], Type[T1], Type[T2]]]
    ) -> ForwardSubclass[Union[Type[T], Type[T1], Type[T2]]]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Optional, Set as TSet, Tuple, Type, TypeVar, overload

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")

class Set(Member[TSet[T], TSet[T]]):
    @overload
    def __new__(cls, item: None = None, default: Optional[set] = None) -> Set[Any]: ...
    @overload
    def __new__(cls, item: Type[T], default: None = None) -> Set[T]: ...
    @overload
    def __new__(cls, item: Tuple[Type[T]], default: None = None) -> Set[T]: ...
    @overload
    def __new__(
        cls, item: Tuple[Type[T], Type[T1]], default: None = None
    ) -> Set[T | T1]: ...
    @overload
    def __new__(
        cls,
        item: Tuple[Type[T], Type[T1], Type[T2]],
        default: None = None,
    ) -> Set[T | T1 | T2]: ...
    @overload
    def __new__(cls, item: Member[T, Any], default: None = None) -> Set[T]: ...
    # With default
    # The splitting is necessary otherwise Mypy type inference fails
    @overload
    def __new__(cls, item: Type[T], default: TSet[T]) -> Set[T]: ...
    @overload
    def __new__(cls, item: Tuple[Type[T]], default: TSet[T]) -> Set[T]: ...
    @overload
    def __new__(
        cls, item: Tuple[Type[T], Type[T1]], default: TSet[T | T1]
    ) -> Set[T | T1]: ...
    @overload
    def __new__(
        cls, item: Tuple[Type[T], Type[T1]], default: TSet[T] | TSet[T1]
    ) -> Set[T | T1]: ...
    @overload
    def __new__(
        cls,
        item: Tuple[Type[T], Type[T1], Type[T2]],
        default: TSet[T | T1 | T2],
    ) -> Set[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        item: Tuple[Type[T], Type[T1], Type[T2]],
        default: TSet[T | T1] | TSet[T | T2] | TSet[T1 | T2],
    ) -> Set[T | T1 | T2]: ...
    @overload
    def __new__(
        cls,
        item: Tuple[Type[T], Type[T1], Type[T2]],
        default: TSet[T] | TSet[T1] | TSet[T2],
    ) -> Set[T | T1 | T2]: ...
    @overload
    def __new__(cls, item: Member[T, Any], default: TSet[T]) -> Set[T]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import (
    Any,
    Callable as TCallable,
    Literal,
    NoReturn,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from .catom import Member

T = TypeVar("T")
T1 = TypeVar("T1")
T2 = TypeVar("T2")
S = TypeVar("S")

class Value(Member[T, T]):
    def __new__(
        cls, default: Any = None, *, factory: Optional[TCallable[[], Any]] = None
    ) -> Value[Any]: ...

class ReadOnly(Member[T, T]):
    @overload
    def __new__(
        cls, kind: None = None, *, default: None = None, factory: None = None
    ) -> ReadOnly[Any]: ...
    @overload
    def __new__(
        cls, kind: None = None, *, default: T, factory: None = None
    ) -> ReadOnly[T]: ...
    @overload
    def __new__(
        cls, kind: None = None, *, default: None = None, factory: TCallable[[], T]
    ) -> ReadOnly[T]: ...
    @overload
    def __new__(
        cls,
        kind: Type[T],
        *,
        default: Optional[T] = None,
        factory: Optional[TCallable[[], T]] = None,
    ) -> ReadOnly[T]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T]],
        *,
        default: Optional[T] = None,
        factory: Optional[TCallable[[], T]] = None,
    ) -> ReadOnly[T]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1]],
        *,
        default: Optional[T | T1] = None,
        factory: Optional[TCallable[[], T | T1]] = None,
    ) -> ReadOnly[T | T1]: ...
    @overload
    def __new__(
        cls,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
        *,
        default: Optional[T | T1 | T2] = None,
        factory: Optional[TCallable[[], T | T1 | T2]] = None,
    ) -> ReadOnly[T | T1 | T2]: ...

class Constant(Member[T, NoReturn]):  # FIXME over-write del ?
    @overload
    def __new__(
        cls, default: None = None, *, factory: None = None, kind: None = None
    ) -> Constant[Any]: ...
    @overload
    def __new__(
        cls, default: None = None, *, factory: TCallable[[], T], kind: None = None
    ) -> Constant[T]: ...
    @overload
    def __new__(
        cls, default: T, *, factory: None = None, kind: None = None
    ) -> Constant[T]: ...
    @overload
    def __new__(
        cls,
        default: Optional[T] = None,
        *,
        factory: Optional[TCallable[[], T]] = None,
        kind: Type[T],
    ) -> Constant[T]: ...
    @overload
    def __new__(
        cls,
        default: Optional[T] = None,
        *,
        factory: Optional[TCallable[[], T]] = None,
        kind: Tuple[Type[T]],
    ) -> Constant[T]: ...
    @overload
    def __new__(
        cls,
        default: Optional[T | T1] = None,
        *,
        factory: Optional[TCallable[[], T | T1]] = None,
        kind: Tuple[Type[T], Type[T1]],
    ) -> Constant[T | T1]: ...
    @overload
    def __new__(
        cls,
        default: Optional[T | T1 | T2] = None,
        *,
        factory: Optional[TCallable[[], T | T1 | T2]] = None,
        kind: Tuple[Type[T], Type[T1], Type[T2]],
    ) -> Constant[T | T1 | T2]: ...

C = TypeVar("C", bound=TCallable)

class Callable(Member[T, T]):
    @overload
    def __new__(
        self, default: None = None, *, factory: TCallable[[], C]
    ) -> Callable[C]: ...
    @overload
    def __new__(self, default: C, *, factory: None = None) -> Callable[C]: ...
    @overload
    def __new__(
        self, default: None = None, *, factory: None = None
    ) -> Callable[TCallable]: ...

class Bool(Member[bool, T]):
    def __new__(
        self, default: bool = False, *, factory: Optional[TCallable[[], bool]] = None
    ) -> Bool[bool]: ...

class Int(Member[int, T]):
    @overload
    def __new__(
        cls,
        default: int = 0,
        *,
        factory: Optional[TCallable[[], int]] = None,
        strict: Literal[True] = True,
    ) -> Int[int]: ...
    @overload
    def __new__(
        cls,
        default: Union[int, float] = 0,
        *,
        factory: Optional[TCallable[[], Union[int, float]]] = None,
        strict: Literal[False],
    ) -> Int[Union[int, float]]: ...

# NOTE this cannot be properly statically typed checked since Mypy will always accept
# an int where a float is expected
class FloatRange(Member[float, T]):
    @overload
    def __new__(
        self,
        low: Optional[float] = None,
        high: Optional[float] = None,
        value: Optional[float] = None,
        *,
        strict: Literal[False] = False,
    ) -> FloatRange[int | float]: ...
    @overload
    def __new__(
        self,
        low: Optional[float] = None,
        high: Optional[float] = None,
        value: Optional[float] = None,
        *,
        strict: Literal[True],
    ) -> FloatRange[float]: ...

class Range(Member[int, T]):
    def __new__(
        self,
        low: Optional[int] = None,
        high: Optional[int] = None,
        value: Optional[int] = None,
    ) -> Range[int]: ...

class Float(Member[float, T]):
    @overload
    def __new__(
        cls,
        default: float = 0.0,
        *,
        factory: TCallable[[], Union[int, float]] = None,
        strict: Literal[False] = False,
    ) -> Float[float]: ...
    @overload
    def __new__(
        cls,
        default: float = 0.0,
        *,
        factory: TCallable[[], float] = None,
        strict: Literal[True],
    ) -> Float[float]: ...  # FIXME we cannot encode that an int will be rejected

class Bytes(Member[bytes, T]):
    @overload
    def __new__(
        cls,
        default: bytes = b"",
        *,
        factory: Optional[TCallable[[], bytes]] = None,
        strict: Literal[True] = True,
    ) -> Bytes[bytes]: ...
    @overload
    def __new__(
        cls,
        default: Union[str, bytes] = b"",
        *,
        factory: Optional[TCallable[[], Union[str, bytes]]] = None,
        strict: Literal[False],
    ) -> Bytes[Union[bytes, str]]: ...

class Str(Member[str, T]):
    @overload
    def __new__(
        cls,
        default: str = "",
        *,
        factory: Optional[TCallable[[], str]] = None,
        strict: Literal[True] = True,
    ) -> Str[str]: ...
    @overload
    def __new__(
        cls,
        default: Union[str, bytes] = "",
        *,
        factory: Optional[TCallable[[], Union[str, bytes]]] = None,
        strict: Literal[False],
    ) -> Str[Union[str, bytes]]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Dict as TDict, Optional, Tuple, Type, TypeVar, overload

from .catom import Member

T = TypeVar("T")

KT = TypeVar("KT")
VT = TypeVar("VT")
KT1 = TypeVar("KT1")
VT1 = TypeVar("VT1")
KT2 = TypeVar("KT2")
VT2 = TypeVar("VT2")

class Dict(Member[TDict[KT, VT], TDict[KT, VT]]):
    # Untyped
    @overload
    def __new__(
        cls, key: None = None, value: None = None, default: Optional[dict] = None
    ) -> Dict[Any, Any]: ...
    # No default
    # Typed keys
    # - type
    @overload
    def __new__(
        cls, key: Type[KT], value: None = None, default: Optional[dict] = None
    ) -> Dict[KT, Any]: ...
    # - 1-tuple
    @overload
    def __new__(
        cls, key: Tuple[Type[KT]], value: None = None, default: Optional[dict] = None
    ) -> Dict[KT, Any]: ...
    # - 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: None = None,
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, Any]: ...
    # - 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: None = None,
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, Any]: ...
    # - member
    @overload
    def __new__(
        cls, key: Member[KT, Any], value: None = None, default: Optional[dict] = None
    ) -> Dict[KT, Any]: ...
    # Typed values
    # - type
    @overload
    def __new__(
        cls, key: None, value: Type[VT], default: Optional[dict] = None
    ) -> Dict[Any, VT]: ...
    # - 1-tuple
    @overload
    def __new__(
        cls, key: None, value: Tuple[Type[VT]], default: Optional[dict] = None
    ) -> Dict[Any, VT]: ...
    # - 2-tuple
    @overload
    def __new__(
        cls,
        key: None,
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[Any, VT | VT1]: ...
    # - 3-tuple
    @overload
    def __new__(
        cls,
        key: None,
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[Any, VT | VT1 | VT2]: ...
    # - member
    @overload
    def __new__(
        cls, key: None, value: Member[VT, Any], default: Optional[dict] = None
    ) -> Dict[Any, VT]: ...
    # Typed value through keyword
    # - type
    @overload
    def __new__(
        cls, key: None = None, *, value: Type[VT], default: Optional[dict] = None
    ) -> Dict[Any, VT]: ...
    # - 1-tuple
    @overload
    def __new__(
        cls, key: None = None, *, value: Tuple[Type[VT]], default: Optional[dict] = None
    ) -> Dict[Any, VT]: ...
    # - 2-tuple
    @overload
    def __new__(
        cls,
        key: None = None,
        *,
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[Any, VT | VT1]: ...
    # - 3-tuple
    @overload
    def __new__(
        cls,
        key: None = None,
        *,
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[Any, VT | VT1 | VT2]: ...
    # - member
    @overload
    def __new__(
        cls,
        key: None = None,
        *,
        value: Member[VT, Any] = None,
        default: Optional[dict] = None,
    ) -> Dict[Any, VT]: ...
    # Typed key and value
    # - value simple type
    #    - key type
    @overload
    def __new__(
        cls, key: Type[KT], value: Type[VT], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...
    #    - key 1-tuple
    @overload
    def __new__(
        cls, key: Tuple[Type[KT]], value: Type[VT], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...
    #    - key 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: Type[VT],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, VT]: ...
    #    - key 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: Type[VT],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, VT]: ...
    #    - key member
    @overload
    def __new__(
        cls, key: Member[KT, Any], value: Type[VT], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...
    # - Value as single element tuple
    #    - key type
    @overload
    def __new__(
        cls, key: Type[KT], value: Tuple[Type[VT]], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...
    #    - key 1-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT]],
        value: Tuple[Type[VT]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT]: ...
    #    - key 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: Tuple[Type[VT]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, VT]: ...
    #    - key 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: Tuple[Type[VT]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, VT]: ...
    #    - key member
    @overload
    def __new__(
        cls,
        key: Member[KT, Any],
        value: Tuple[Type[VT]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT]: ...
    # - Value as 2-tuple
    #    - key type
    @overload
    def __new__(
        cls,
        key: Type[KT],
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1]: ...
    #    - key 1-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT]],
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1]: ...
    #    - key 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, VT | VT1]: ...
    #    - key 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, VT | VT1]: ...
    #    - key member
    @overload
    def __new__(
        cls,
        key: Member[KT, Any],
        value: Tuple[Type[VT], Type[VT1]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1]: ...
    # - Value as 3-tuple
    #   - key type
    @overload
    def __new__(
        cls,
        key: Type[KT],
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1 | VT2]: ...
    #   - key 1-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT]],
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1 | VT2]: ...
    #   - key 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, VT | VT1 | VT2]: ...
    #   - key 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, VT | VT1 | VT2]: ...
    #   - key member
    @overload
    def __new__(
        cls,
        key: Member[KT, Any],
        value: Tuple[Type[VT], Type[VT1], Type[VT2]],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT | VT1 | VT2]: ...
    # - value as member
    #   - key type
    @overload
    def __new__(
        cls, key: Type[KT], value: Member[VT, Any], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...
    #   - key 1-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT]],
        value: Member[VT, Any],
        default: Optional[dict] = None,
    ) -> Dict[KT, VT]: ...
    #   - key 2-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1]],
        value: Member[VT, Any],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1, VT]: ...
    #   - key 3-tuple
    @overload
    def __new__(
        cls,
        key: Tuple[Type[KT], Type[KT1], Type[KT2]],
        value: Member[VT, VT],
        default: Optional[dict] = None,
    ) -> Dict[KT | KT1 | KT2, VT]: ...
    #   - key member
    @overload
    def __new__(
        cls, key: Member[KT, KT], value: Member[VT, VT], default: Optional[dict] = None
    ) -> Dict[KT, VT]: ...



# --------------------------------------------------------------------------------------
# Copyright (c) 2013-2021, Nucleic Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# --------------------------------------------------------------------------------------
from typing import Any, Tuple, TypeVar, Union, overload

from .catom import Member

# We cannot type properly an enum, as it would require some dynamic literal so this is
# just a best effort.

T = TypeVar("T")
T1 = TypeVar("T1")

class Enum(Member[T, T]):
    @overload
    def __new__(cls, item: T) -> Enum[T]: ...  # This is hacky but should do
    @overload
    def __new__(cls, *items: T) -> Enum[T]: ...
    @property
    def items(self) -> Tuple[T, ...]: ...
    def added(self: Enum[T], *items: T1) -> Enum[Union[T, T1]]: ...
    def removed(self: Enum[T], *items: Any) -> Enum[T]: ...
    def __call__(self: Enum[T], item: T1) -> Enum[Union[T, T1]]: ...
